import hashlib as hl
from io import TextIOWrapper
from pws256 import PwsType
from pws256.src import _Password


class User:
    def __init__(self, username: str, password: PwsType):
        self.username = username
        if not issubclass(type(password), _Password):
            raise TypeError(
                f"Password must be derived from _Password (eg. Password() or defaultpass())"
            )
        self.password = password

    @classmethod
    def load_from_file(cls, username: str, file: TextIOWrapper | str, hsh_func=hl.sha256, hsh_enter: str | bytes = bytes, hsh_after = ".hexdigest()", close=False):
        if isinstance(file, str):
            file = open(file)

        for line in file.readlines():
            line = line[:-1]
            if line.split(",")[0] == username:
                return User(username, _Password(
                    hashed = line.split(",")[1],
                    hsh_func = hsh_func,
                    hsh_enter = hsh_enter,
                    hsh_after = hsh_after,
                    salt = line.split(",")[2]
                ))
            
        if close:
            file.close()
        
    def save_to_file(self, file: TextIOWrapper | str, close=False):
        if isinstance(file, str):
            file = open(file, "a")
        file.write(f"{self.username},{self.password.hashed},{self.password.get_salt()}\n")
        if close:
            file.close()

    def to_dict(self):
        return {
            "username": self.username,
            "password": self.password.to_dict()
        }
        



class Validator:
    def __init__(self, hashed, hsh_func, hsh_enter, hsh_after, salt):
        self.hashed, self.__hsh_func, self.__hsh_enter, self.__hsh_after, self._salt = hashed, hsh_func, hsh_enter, hsh_after, salt
    
    def hash_password_validator(self, raw):
        return hash_password(raw, self.__hsh_func, self.__hsh_enter, self.__hsh_after, self._salt)

    def validate(self, other):
        return self.hash_password_validator(other) == self.hashed
    



def hash_password(raw, hsh_func, hsh_enter, hsh_after, salt):
    encoded = hsh_func(((salt + raw) if hsh_enter == str else (salt + raw).encode()))
    if hsh_after:
        encoded = eval("encoded" + hsh_after, dict({"encoded": encoded}))

    return encoded


"""
This is used to verify passwords with custom hash functions (or sha256 by default)

--- pws256/

>>> import pws256
>>> pw = pws256.Password("password")
>>> pw.verify("password")
True
>>> pw.verify("pasword")
False
>>> pw.hsh_func
<built-in function openssl_sha256>
>>> pw2 = pws256.Password(
...    raw = "password",
...    hsh_func = lambda x : "".join(reversed(x)),
...    hsh_enter = str,
...    hsh_after = None
...)
...
>>> pw2.verify("password")
True
>>> pw2.hashed
"drowssap"
>>>
"""

import hashlib as hl
import secrets
import types
from typing import TypeVar
from pws256.src.base import Validator, hash_password


__all__ = [
    "_Password",
    "Password",
    "defaultpass",
    "PwsType"
]

class CantSetError(Exception): ...

class _Password(Validator):
    def __init__(self, hashed, hsh_func, hsh_enter, hsh_after, salt):
        super().__init__(hashed, hsh_func, hsh_enter, hsh_after, salt)
        self.__salt_settable = False

    @property
    def hsh_func(self):
        return self.__hsh_func
    
    @hsh_func.setter
    def hsh_func(self, obj):
        if not isinstance(obj, types.FunctionType):
            raise TypeError(
                f"obj must be a function, not {obj.__class__.__name__}"
            )
        self.__hsh_func = obj

    @property
    def salt(self):
        return self._salt
    
    def get_salt(self):
        return self.salt
    
    @salt.setter
    def salt(self, val):
        if not self.__salt_settable:
            raise CantSetError(
                "You can't set the salt value"
            )
        else:
            self._salt = val
    
    @property
    def hsh_after(self):
        return self.__hsh_after
    
    @hsh_after.setter
    def hsh_after(self, obj):
        if not isinstance(obj, str):
            raise TypeError(
                f"obj must be of type 'str', not '{obj.__class__.__name__}'"
            )
        
    def salt_settable(self, which: bool = None):
        if which != None:
            if not isinstance(which, bool):
                raise TypeError(
                    f"which must be of type bool, not {which.__class__.__name__}"
                )
            self.__salt_settable = which
        return self.__salt_settable
    
    def __repr__(self):
        hsh_after = "" if self.__hsh_after == None else self.__hsh_after
        return f"<<PWS256 PASSWORD: {self.__hsh_func.__name__}(obj: {self.__hsh_enter.__name__}){hsh_after}>>"
    
    def to_dict(self):
        return {
            "hashed": self.hashed,
            "hsh_func": "sha256" if self._Validator__hsh_func == hl.sha256 else "sha512",  # 2 Choices for the shell version
            "hsh_enter": "str" if self._Validator__hsh_enter == str else "bytes",
            "hsh_after": self._Validator__hsh_after,
            "salt": self._salt
        }
    


class Password(_Password):
    def __init__(self, raw: str, hsh_func=hl.sha256, hsh_enter: str | bytes = bytes, hsh_after = ".hexdigest()"):
        salt = secrets.token_hex(32)
        encoded = hash_password(raw, hsh_func, hsh_enter, hsh_after, salt)


        super().__init__(encoded, hsh_func, hsh_enter, hsh_after, salt) # Initialise _Password with the result of the password

    

class defaultpass(Password):
    def __init__(self, raw: str):
        super().__init__(raw, hl.sha256, bytes, ".hexdigest()")


PwsType = TypeVar("PwsType", Password, _Password)


if __name__ == "__main__":
    print("Password: \"hello\" with sha256\n\n")

    pw = defaultpass("hello") # Generates internal code of sha256("hello".encode()).hexdigest()

    print("Password is hello!" if pw.validate("hello") else "uh oh what did u do wrong")
    print("\n")
    print("Password: \"hello\" with reversed function\n\n")
    
    def reverse(x):
        return "".join(x)
    
    pw2 = Password("hello", hsh_func = reverse, hsh_enter = str, hsh_after = None) # Generates internal code of reverse("hello")

    print("Password is hello!" if pw2.validate("hello") else "uh oh what did u do wrong")
    


import numpy

DEFAULT_N = 4330967
DEFAULT_T = 4318800

def coprime(i, t):
    return numpy.gcd(i, t) == 1

class PublicKey:
    def __init__(self, n, t):
        self.n = n
        self.t = t

    def get(self):
        for i in range(2, self.t):
            if coprime(i, self.t):
                return i
            

class PrivateKey:
    def __init__(self, e: PublicKey):
        self.e = e
        self.n = e.n
        self.t = e.t
        self.s = self.create_key()

    def create_key(self):
        e = self.e.get()
        d = e
        while True:
            d += 1
            if (e * d) % self.t == 1:
                return d
            


class KeyPair:
    def __init__(self, n = DEFAULT_N, t = DEFAULT_T):
        self.e = PublicKey(n, t)
        self.d = PrivateKey(self.e)

    def encrypt(self, msg):
        c = pow(msg, self.e.get()) % self.e.n
        return c
    
    def decrypt(self, c):
        m = pow(c, self.d.s) % self.e.n
        return m
    




import os

def setupchecks():
    path = os.path.dirname(__file__) + "/srcimport.py"
    with open(path, "w") as file:
        file.write("""\"THIS FILE IS PREGENERATED. ANY CHANGE YOU MAKE WILL BE OVERWRITTEN\"

from pws256.src import _Password, Password, PwsType, defaultpass""")
        
    path = os.path.dirname(__file__) + "/usersrcimport.py"
    with open(path, "w") as file:
        file.write("""\"THIS FILE IS PREGENERATED. ANY CHANGE YOU MAKE WILL BE OVERWRITTEN\"

from pws256.src.users import User""")
        
    path = os.path.dirname(__file__) + "/rsasrcimport.py"
    with open(path, "w") as file:
        file.write("""\"THIS FILE IS PREGENERATED. ANY CHANGE YOU MAKE WILL BE OVERWRITTEN\"

from pws256.src.rsa import PublicKey, PrivateKey, KeyPair""")
        




"""
Module used for custom user classes to write in files

--- pws256/users

>>> import pws256 as pws
>>> from pws256.users import User
>>> pw = pws.Password("hallo")
>>> u = User("sigma123", pw)
>>> u.save_to_file("test.csv")
>>> usr = User.load_from_file("sigma123", "test.csv")
>>> usr.password.validate("hallo")
True
"""
from pws256._setup.setup import setupchecks

setupchecks()
del setupchecks
from pws256._setup.usersrcimport import *


"""
This is used to verify passwords with custom hash functions (or sha256 by default)

--- pws256/

>>> import pws256
>>> pw = pws256.Password("password")
>>> pw.verify("password")
True
>>> pw.verify("pasword")
False
>>> pw.hsh_func
<built-in function openssl_sha256>
>>> pw2 = pws256.Password(
...    raw = "password",
...    hsh_func = lambda x : "".join(reversed(x)),
...    hsh_enter = str,
...    hsh_after = None
...)
...
>>> pw2.verify("password")
True
>>> pw2.hashed
"drowssap"
>>>
"""

__version__ = "1.1.14"

import os, datetime
try:
    if os.environ["PWS256-VERBOSE"] == "3":
        now = datetime.datetime.now().strftime("%H:%M:%S")
        print(f"[{now}] pws256 start")
except:
    pass


from pws256._setup.setup import setupchecks

setupchecks()
del setupchecks, datetime, os
try:
    del now
except:
    pass

from pws256._setup.srcimport import *


from setuptools import setup, find_packages
with open("pws256/README.md") as file:
    setup(
        name='pws256',
        version='1.1.14',
        description="pws256: A module that is used for users and passwords, and also has validation methods to check passwords with strings",
        long_description=file.read(),
        long_description_content_type='text/markdown',
        packages=["pws256", "pws256.src", "pws256._setup"],
        author='Hector Robertson',
        py_modules=["rsa", "users"],
        install_requires=["numpy"],
        readme="pws256/README.md"
    )